/**************************************
 * $Author: kaneda $
 * $Date: 2003/03/12 12:03:51 $
 **************************************/

/* ph_conn.h */

#ifndef __PH_CONN_H__
#define __PH_CONN_H__

#include "ph_util.h"
#include "ph_addr.h"


/* Macro Values */
enum {
  PH_FD_TYPE_READ		= 0,
  PH_FD_TYPE_WRITE		= 1,
  PH_FD_TYPE_ERROR 		= 2,

  PH_CONNECT_TIMEOUT		= 5,
  PH_SSH_TIMEOUT		= 10
};


/* perform dominating-set based routing */
//#define		PH_ROUTING_ENABLE_DOMINATING

/* Structures */


enum ph_conn_state {
  connecting, accepting, opened, closed
};

typedef enum ph_conn_state ph_conn_state_t;

/* ph_tcp_attr_t is ... */
struct ph_tcp_attr {
  int fd;				/* socket filed descripter */  
};

typedef struct ph_tcp_attr 	* ph_tcp_attr_t;


/* ph_ssh_attr_t is ... */
struct ph_ssh_attr {
  int send_fd, recv_fd;  			
  int pid;  				/* ssh process's id */
};

typedef struct ph_ssh_attr 	* ph_ssh_attr_t;



/* ph_conn_t is ... */
struct ph_conn {
  ph_addr_t		addr;		/* destination address of this connection */
  ph_conn_state_t	state;		/* */
  
  void * 		attr;		/* ph_tcp_conn_t or ph_ssh_conn_t */

  int			is_authed;	/* authentication flag */
  int			is_repaired;	/* for routing */

  void * (*connect_func_p)(ph_addr_t);
  void * (*listen_func_p)(ph_addr_t);
  void * (*accept_func_p)(void *, struct sockaddr *, socklen_t *);
  ssize_t (*send_func_p)(void * , void *, size_t);
  ssize_t (*recv_func_p)(void *, void * buf, size_t);
  void (*close_func_p)(void *);

  void (*dispose_func_p)(void *);
  int (*fd_isset_func_p)(void *, fd_set *, int);
  void (*fd_set_func_p)(void *, fd_set *, int);

#ifdef PH_ROUTING_ENABLE_DOMINATING
  int 			is_permanent;
#endif

  pthread_mutex_t	mp;		/* to synchronize access to variables */
  int			nref;		/* the number of refereces */
};

typedef struct ph_conn 	* ph_conn_t;


PH_DEFINE_LIST_STRUCTURE(ph_conn) 


/* Public Function Prototypes */

/* Following functions are defined in conn.c */
ph_conn_t ph_conn_new();
void ph_conn_dispose(ph_conn_t conn);

ph_conn_t ph_conn_connect(ph_addr_t addr);
int ph_conn_connecting(ph_conn_t conn);
ph_conn_t ph_conn_listen(int available_ports[2], int * listen_port_p);
ph_conn_t ph_conn_accept(ph_conn_t listen_conn);

ssize_t ph_conn_send(ph_conn_t conn, void * buf, size_t len);
ssize_t ph_conn_recv(ph_conn_t conn, void * buf, size_t len);
void ph_conn_close(ph_conn_t conn);

void ph_conn_lock(ph_conn_t conn);
void ph_conn_unlock(ph_conn_t conn);

ph_conn_state_t	ph_conn_get_state(ph_conn_t conn);
void ph_conn_set_state(ph_conn_t conn, ph_conn_state_t state);
ph_addr_type_t ph_conn_get_addr_type(ph_conn_t conn);
void ph_conn_set_addr_type(ph_conn_t conn, ph_addr_type_t type);
#ifdef PH_ROUTING_ENABLE_DOMINATING
int ph_conn_get_is_permanent(ph_conn_t conn);
void ph_conn_set_is_permanent(ph_conn_t conn, int is_permanent);
#endif

PH_DEFINE_LIST_FUNC_PROTOTYPE(ph_conn) 


void ph_conn_list_print(ph_conn_list_t conn_list);
ph_conn_t ph_conn_list_find_same_addr(ph_conn_list_t conn_list, ph_addr_t addr);
ph_conn_cell_t ph_conn_list_init_loop(ph_conn_list_t conn_list);
ph_conn_cell_t ph_conn_list_iterate(ph_conn_list_t conn_list, ph_conn_cell_t p);



/* Following functions are defined in tcp.c */
ph_tcp_attr_t ph_tcp_attr_new(int fd);
void ph_tcp_attr_dispose(void * p);
void * ph_tcp_attr_connect(ph_addr_t addr);
void * ph_tcp_attr_listen(ph_addr_t addr);
void * ph_tcp_attr_accept(void * p, struct sockaddr * cliaddr_p, socklen_t * len_p);
ssize_t ph_tcp_attr_send(void * p, void * buf, size_t len);
ssize_t ph_tcp_attr_recv(void * p, void * buf, size_t len);
void ph_tcp_attr_close(void * p);
int ph_tcp_attr_fd_isset(void * p, fd_set * fdset_p, int type);
void ph_tcp_attr_fd_set(void * p, fd_set * fd_set_p, int type);


/* Following functions are defined in ssh.c */
ph_ssh_attr_t ph_ssh_attr_new(int send_fd, int recv_fd, int pid);
void ph_ssh_attr_dispose(void * p);
void * ph_ssh_attr_connect(ph_addr_t addr);
void * ph_ssh_attr_listen(ph_addr_t addr);
void * ph_ssh_attr_accept(void * p, struct sockaddr * cliaddr_p, socklen_t * len_p);
ssize_t ph_ssh_attr_send(void * p, void * buf, size_t len);
ssize_t ph_ssh_attr_recv(void * p, void * buf, size_t len);
void ph_ssh_attr_close(void * p);
int ph_ssh_attr_fd_isset(void * p, fd_set * fdset_p, int type);
void ph_ssh_attr_fd_set(void * p, fd_set * fdset_p, int type);

#endif







