/**************************************
 * $Author: kaneda $
 * $Date: 2003/03/12 09:04:35 $
 **************************************/

/* interface.c */

#include "phoenix.h"

static ph_env_t env = NULL;



static void sig_pipe(int signo) {
  //err_ret("BROKEN PIPE");
  return;
}

#ifdef __PH_WIN32__
static void initialize_winsock2() {
  WORD		wVersionRequested = MAKEWORD(2, 2);
  WSADATA	wsaData;
    
  WSAStartup(wVersionRequested, &wsaData);  
}
#endif

#ifndef __PH_WIN32__
static void initialize_misc() {
  {
    sigset_t set;

    sigemptyset(&set);
    sigaddset(&set, SIGALRM);
    ph_pthread_sigmask(SIG_BLOCK, &set, NULL);
  } 

  ph_signal(SIGPIPE, sig_pipe);

  {
#if 0
    /* tau : probably, what we need to is to call seed48 */
    int i;
    unsigned short param[7];
    
    for (i = 0; i < 7; i++) {
      struct timeval t;
      gettimeofday(&t, NULL);
      param[i] = (unsigned short)t.tv_usec;
    }
    lcong48(param);
#else
    int i;
    unsigned short seed[3];
    for (i = 0; i < 3; i++) {
      struct timeval t;
      gettimeofday(&t, NULL);
      seed[i] = (unsigned short)t.tv_usec;
    }
    seed48(seed);
#endif
  }

}
#endif



static void ph_initialize_sub(int listen_port_lower, int listen_port_upper) {
  ph_debug_print("phoenix_initialize()\t\t: \n");
  
#ifdef __PH_WIN32__
  initialize_winsock2();
#endif

#ifndef __PH_WIN32__
  initialize_misc();
#endif
   
  env = ph_env_new(listen_port_lower, listen_port_upper);
  ph_pt2pt_initialize(env->pt2pt);
#if 0
  { 
    ph_vp_set_t vp_set;
    vp_set = ph_vp_set_make_singleton(env->pt2pt->resource_vp);
    ph_pt2pt_assume_vp_set(env->pt2pt, vp_set);
    ph_vp_set_dispose(vp_set);
  }
#endif
}

/* spawn phoenix daemon and initialize variables (this function is called when a program starts) */
void ph_initialize(int listen_port) {
  ph_initialize_sub(listen_port, listen_port);
}

void ph_initialize2(int listen_port_lower, int listen_port_upper) {
  ph_initialize_sub(listen_port_lower, listen_port_upper);
}


/* assume that this vp_set is allocated to it. */
void ph_assume_vps(ph_vps_t vp_set) {
  assert(env != NULL);
  assert(vp_set != NULL);
  ph_pt2pt_assume_vp_set(env->pt2pt, (ph_vp_set_t)vp_set);
}


/* release vp_set that is allocated to it. */
void ph_release_vps(ph_vps_t vp_set) {
  assert(env != NULL);
  assert(vp_set != NULL);
  ph_pt2pt_release_vp_set(env->pt2pt, (ph_vp_set_t)vp_set);
}

ph_vps_t ph_get_assumed_vps() {
  assert(env != NULL);
  return (ph_vp_set_t)ph_pt2pt_get_assumed_vps(env->pt2pt);
}

ph_vp_t ph_get_resource_vp() {
  assert(env != NULL);
  return env->pt2pt->resource_vp;
}

/* declare that this host should be a neighbor node to which a direct connections can be initiated */
void ph_add_contact_point(char * addr, int port, ph_addr_type_t addr_type, ...) {
  va_list ap;
  char service[PH_BUFSIZE];


  assert(env != NULL);

  ph_mzero(service, PH_BUFSIZE);
  ph_snprintf(service, PH_BUFSIZE, "%d", port);   

  va_start(ap, addr_type);
  switch (addr_type) {
  case tcp:
    ph_pt2pt_add_contact_point(env->pt2pt, addr, service, addr_type, NULL);
    break;
  case ssh:
    ph_pt2pt_add_contact_point(env->pt2pt, addr, service, addr_type, (char *)va_arg(ap, char *));
    break;
  default:
    ph_err_quit("ph_add_contact_point() : no matching");
    break;
  }
  va_end(ap);
}



/* send message to vp */
void ph_send(ph_vp_t vp, char * body, size_t len, int tag) {
  assert(env != NULL);
  ph_pt2pt_send_user_msg(env->pt2pt, vp, body, len, tag);
}


/* receive message */
ssize_t ph_recv(char ** body, int tag) {
  ph_msg_t msg;
  ssize_t len;

  assert(env != NULL);

  msg = ph_pt2pt_recv_user_msg(env->pt2pt, tag, 1);
  len = msg->len;
  (*body) = msg->body; 
  msg->body = NULL;
  ph_msg_dispose(msg);
  return len;
}

void ph_invalidate_routing() {
  assert(env != NULL);
  ph_pt2pt_invalidate_routing(env->pt2pt);
}



/* finish (this function is called when a program exits) */
void ph_finalize() {
  assert(env != NULL);
  ph_pt2pt_finalize(env->pt2pt);
  
#ifdef __PH_WIN32__
  WSACleanup();
#endif

  ph_env_dispose(env);
}


int ph_messages_is_empty() {
   assert(env != NULL);
   return ph_pt2pt_messages_is_empty(env->pt2pt);
}

void ph_load_messages(char * filename) {
  assert(env != NULL);
  assert(filename != NULL);
  ph_pt2pt_load_messages(env->pt2pt, filename);
}

void ph_store_messages(char * filename) {
  assert(env != NULL);
  assert(filename != NULL);
  ph_pt2pt_store_messages(env->pt2pt, filename);
}

ph_vp_t ph_get_random_vp() {  
  ph_vp_t vp;

  vp = ((long long)(lrand48()) << 31) + (long long)(lrand48());
  if (vp >= PHOENIX_VP_LIMIT) {
    vp = PHOENIX_MAX_VP;
  }
  assert((PH_MIN_VP <= vp) && (vp <= PHOENIX_MAX_VP));
  return vp;
}


ph_vp_t ph_get_temp_vp() {
  return ph_pt2pt_get_temp_vp();
}


ssize_t ph_try_recv(char ** body, int tag) {
  ph_msg_t msg;
  ssize_t len;

  assert(env != NULL);

  if ((msg = ph_pt2pt_recv_user_msg(env->pt2pt, tag, 0)) == NULL) {
    (*body) = NULL;
    return 0;
  }

  len = msg->len;
  (*body) = msg->body; 
  msg->body = NULL;
  ph_msg_dispose(msg);
  return len;
}


void ph_get_routes(ph_vps_t * vp_set_p) {
  /*
  ph_routing_t routing;
  ph_routing_entry_cell_t p;

  assert(env != NULL);
  assert(env->pt2pt != NULL);

  routing = env->pt2pt->routing;

  assert(env->pt2pt->routing != NULL); 
  ph_pthread_mutex_lock(&routing->mp);
  for (p = routing->entry_list->head; p != NULL; p = p->next) {
    if (p->self->metric < PH_MAX_METRIC) {
      ph_vp_set_list_insert(vp_set_list, ph_vp_set_dup(p->self->dest));
    }
  }
  ph_pthread_mutex_unlock(&routing->mp);
  */
}

ph_pt2pt_t ph_get_pt2pt() {
  return env->pt2pt;
}


ph_vps_t ph_vps_dup(ph_vps_t vps) {
  return (ph_vps_t)ph_vp_set_dup((ph_vp_set_t)vps);
}


ph_vps_t ph_vps_make_empty() {
  return (ph_vps_t)ph_vp_set_make_empty();
}


ph_vps_t ph_vps_make_singleton(ph_vp_t vp) {
  assert((PH_MIN_VP <= vp) && (vp <= PH_MAX_VP));
  return (ph_vps_t)ph_vp_set_make_singleton(vp);
}


ph_vps_t ph_vps_make_range(ph_vp_t l, ph_vp_t u) {
  assert((PH_MIN_VP <= l) && (l <= PH_MAX_VP));
  assert((PH_MIN_VP <= u) && (u <= PH_MAX_VP));

  return (ph_vps_t)ph_vp_set_make_range(l, u - ph_get_ph_vp_t(1));
}


void ph_vps_dispose(ph_vps_t vps) {
  ph_vp_set_dispose((ph_vp_set_t)vps);
}
  

ph_vps_t ph_vps_make_union(ph_vps_t p, ph_vps_t q) {
  return (ph_vps_t)ph_vp_set_make_union((ph_vp_set_t)p, (ph_vp_set_t)q);
}


ph_vps_t ph_vps_make_diff(ph_vps_t p, ph_vps_t q) {
  return (ph_vps_t)ph_vp_set_make_diff((ph_vp_set_t)p, (ph_vp_set_t)q);
}


ph_vp_t ph_vps_get_min_elem(ph_vps_t vps) {
  return ph_vp_set_get_min_elem((ph_vp_set_t)vps);
}


ph_vp_t ph_vps_get_max_elem(ph_vps_t vps) {
  return ph_vp_set_get_max_elem((ph_vp_set_t)vps);
}


ph_vp_t ph_vps_get_any_elem(ph_vps_t vps) {
  return ph_vp_set_get_any_elem((ph_vp_set_t)vps);
}
  
  
int ph_vps_is_empty(ph_vps_t vps) {
  return ph_vp_set_is_empty((ph_vp_set_t)vps);
}


int ph_vps_is_sub(ph_vps_t p, ph_vps_t q) {
  return ph_vp_set_is_sub((ph_vp_set_t)p, (ph_vp_set_t)q);
}

int ph_vps_is_equal(ph_vps_t p, ph_vps_t q) {
  return ph_vp_set_is_equal((ph_vp_set_t)p, (ph_vp_set_t)q);
}

int ph_vps_is_exclusive(ph_vps_t p, ph_vps_t q) {
  return ph_vp_set_is_exclusive((ph_vp_set_t)p, (ph_vp_set_t)q);
}


int ph_vps_is_member(ph_vps_t vps, ph_vp_t vp) {
  return ph_vp_set_is_member((ph_vp_set_t)vps, vp);
}



#ifdef PH_VIEWER

void ph_viewer_initialize(int viewer_update_interval, char * addr, int port, ph_addr_type_t addr_type, ...) {
  va_list ap;
  char service[PH_BUFSIZE];

  assert(env != NULL);
  ph_mzero(service, PH_BUFSIZE);
  ph_snprintf(service, PH_BUFSIZE, "%d", port);   

  va_start(ap, addr_type);
  switch (addr_type) {
  case tcp:
    ph_pt2pt_viewer_initialize(env->pt2pt, viewer_update_interval, addr, service, addr_type, NULL);
    break;
  case ssh:
    ph_pt2pt_viewer_initialize(env->pt2pt, viewer_update_interval, addr, service, addr_type, (char *)va_arg(ap, char *));
    break;
  default:
    ph_err_quit("ph_viewer_initialize() : no matching");
    break;
  }
  va_end(ap);
}

 
void ph_viewer_send(void * body, size_t len) {
  assert(env != NULL);
  assert(body != NULL);
  ph_pt2pt_viewer_put_comment(env->pt2pt, body, len);
}

#endif








