/**************************************
 * $Author: kaneda $
 * $Date: 2003/02/12 12:45:20 $
 **************************************/

/* handler.c */

#include "ph_pt2pt.h"


static void insert_conn(ph_pt2pt_t pt2pt, ph_conn_t conn) {
  if (conn == NULL) {
    /* ph_conn_connect() or ph_conn_accpet() fails */
    /*
#ifdef PH_ROUTING_ENABLE_DOMINATING
    ph_routing_neigh_t neigh;
    ph_pthread_mutex_lock(&pt2pt->routing->mp);
    while ((neigh = ph_routing_neigh_list_find_same_addr(pt2pt->routing->neigh_list, conn->addr)) != NULL) {
      ph_routing_neigh_list_delete(pt2pt->routing->neigh_list, neigh);
    }
    ph_pthread_mutex_unlock(&pt2pt->routing->mp);
#endif
    */
    return;
  }

  assert(ph_conn_get_state(conn) == opened);

#ifndef PH_ROUTING_ENABLE_DOMINATING
  /* Do not send routing update message if it is not connection for MST */
  ph_routing_event_list_enqueue(pt2pt->routing->event_list, ph_routing_event_new(repair, conn, NULL)); 
#else 

  {
    ph_vp_set_t vps[2], local_vps;
    ph_routing_neigh_cell_t q;
    
    local_vps = ph_vp_set_sync_dup(pt2pt->vp_set);
    
    ph_pthread_mutex_lock(&pt2pt->routing->mp);
    ph_pthread_mutex_lock(&pt2pt->conn_list->mp);
    
    vps[tcp] = ph_vp_set_make_empty();
  vps[ssh] = ph_vp_set_make_empty();
  for (q = pt2pt->routing->neigh_list->head; q != NULL; q = q->next) {
    ph_vp_set_t tmp;

    if (q->self->conn == NULL) {
      continue;
    }
    switch (ph_conn_get_addr_type(q->self->conn)) {
    case tcp:
    case unknown:
      tmp = ph_vp_set_make_union(vps[tcp], q->self->local);
      ph_vp_set_dispose(vps[tcp]);
      vps[tcp] = tmp;
      break;
    case ssh:
      tmp = ph_vp_set_make_union(vps[ssh], q->self->local);
      ph_vp_set_dispose(vps[ssh]);
      vps[ssh] = tmp;
      break;
    default:
      ph_err_quit("connector_check() : no matching");
      break;
    }
  }
  ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
  ph_pthread_mutex_unlock(&pt2pt->routing->mp);


    ph_pt2pt_handle_msg(pt2pt, ph_pt2pt_create_meta_msg(pt2pt, conn, conn_type, ph_conn_get_addr_type(conn), local_vps, vps[tcp], vps[ssh]));

    ph_vp_set_dispose(local_vps);
    ph_vp_set_dispose(vps[tcp]);
    ph_vp_set_dispose(vps[ssh]);   
  }

#endif

  ph_conn_list_insert(pt2pt->conn_list, conn);
}


static void delete_conn(ph_pt2pt_t pt2pt, ph_conn_t conn) {
#ifndef PH_ROUTING_ENABLE_DOMINATING
  /* Do not send routing update message if it is not connection for MST */
  ph_routing_event_list_enqueue(pt2pt->routing->event_list, ph_routing_event_new(fail, conn, ph_routing_entry_list_new(0)));
#endif
  ph_conn_set_state(conn, closed);
  ph_msg_list_clear_conn(pt2pt->msg_list, conn); // TODO
}


void ph_pt2pt_handle_connect(ph_pt2pt_t pt2pt, ph_addr_t addr) {
  ph_conn_t conn;

  conn = ph_conn_connect(addr);
  //ph_auth_send_cookie(pt2pt, conn);
  insert_conn(pt2pt, conn);		 /* TODO: support non-block connect() */
}


void ph_pt2pt_handle_connecting(ph_pt2pt_t pt2pt, ph_conn_t conn, fd_set * r_fdset_p, fd_set * w_fdset_p) {
  assert(0);

  /* TODO
  if (ph_conn_connecting(conn) == 0) {
    return;
  }
  
  assert(ph_conn_get_state(conn) == opened);
  insert_conn(pt2pt, conn);
  */
}


void ph_pt2pt_handle_accepting(ph_pt2pt_t pt2pt, ph_conn_t conn, fd_set * r_fdset_p) {
  if ((*conn->fd_isset_func_p)(conn->attr, r_fdset_p, PH_FD_TYPE_READ) == 0) {
    return;
  }

  insert_conn(pt2pt, ph_conn_accept(conn));
}


void ph_pt2pt_handle_opened(ph_pt2pt_t pt2pt, ph_conn_t conn, fd_set * r_fdset_p) {
  ph_msg_t msg;

  if ((*conn->fd_isset_func_p)(conn->attr, r_fdset_p, PH_FD_TYPE_READ) == 0) {
    return;
  }
  
  if ((msg = ph_msg_recv(conn)) == NULL) {
    // ph_err_msg("recv failure");
    delete_conn(pt2pt, conn);
    return;
  }
  
  ph_pt2pt_handle_msg(pt2pt, msg);
}


void ph_pt2pt_handle_error(ph_pt2pt_t pt2pt, ph_conn_t conn, fd_set * e_fdset_p) {
  if ((*conn->fd_isset_func_p)(conn->attr, e_fdset_p, PH_FD_TYPE_ERROR) == 0) {
    return;
  }

  delete_conn(pt2pt, conn);
}


void ph_pt2pt_handle_msg(ph_pt2pt_t pt2pt, ph_msg_t msg) {
  assert(msg != NULL);

  switch (msg->type) {
  case meta:
    if (ph_msg_is_local_meta_msg(msg, pt2pt->vp_set)) {
      ph_pt2pt_recv_meta_msg(pt2pt, msg);
    } else {
      //ph_pt2pt_send_meta_msg(pt2pt, msg);
      ph_msg_list_enqueue(pt2pt->msg_list, msg);
    }
    break;
  case user:
    ph_routing_route_enqueue_msg(pt2pt->routing, pt2pt->msg_list, msg);
    break;
  default:
    ph_err_quit("ph_pt2pt_handle_msg() : no matching");
    break;
  }
}








