/**************************************
 * $Author: kaneda $
 * $Date: 2003/03/12 09:04:36 $
 **************************************/

/* meta.c */

#include "ph_pt2pt.h"


static char * create_meta_body_routing(ph_pt2pt_t pt2pt, va_list ap, size_t * len_p) {
  char * body;
  char * entries;
  size_t len;
  
  entries = (char *)va_arg(ap, char *);
  len = (size_t)va_arg(ap, size_t);
  body = ph_mdup(entries, len);
  (*len_p) = len;
  return body;
}


static void create_meta_body_view_sub0(ph_pt2pt_t pt2pt, char * body) {
  gethostname(body, 100);
  snprintf(body+100, 100, "%d", pt2pt->listen_port);
  
  ph_pthread_mutex_lock(&pt2pt->mp);
  //ph_mmove(body+200, pt2pt->viewer_comment, 100);
#ifdef PH_ROUTING_ENABLE_DOMINATING
  snprintf(body+200, 100, "%s", (pt2pt->routing->is_gateway) ? "G" : "N");
#else 
  snprintf(body+200, 100, "N");
#endif
  ph_pthread_mutex_unlock(&pt2pt->mp);

  snprintf(body + 300, 100, "%d %d", pt2pt->routing->entry_list->nelm, pt2pt->msg_list->nelm);
}

static int get_conn_no(ph_pt2pt_t pt2pt, ph_conn_t conn) {
  if (conn == NULL) {
    return 0;
  } else {
    int i;
    ph_conn_cell_t p;
    
    for (p = pt2pt->conn_list->head, i = 1; p != NULL; p = p->next, i++) {
      if (p->self == conn) {
	return i;
      }
    }
  }
  return 0;
}

static void create_meta_body_view_sub1(ph_pt2pt_t pt2pt, char * body) {
  int PH_ENTRY_SIZE = 100;
  ph_routing_entry_cell_t p;
  char * s;

  for (p = pt2pt->routing->entry_list->head, s = body; p != NULL; p = p->next, s += PH_ENTRY_SIZE) {
     snprintf(s, PH_ENTRY_SIZE, PH_LLD " " PH_LLD " %d %d %d", 
	      p->self->dest->lower, p->self->dest->upper, get_conn_no(pt2pt, p->self->conn), p->self->metric, 
	      (p->self->conn != NULL) ? (int)p->self->conn->addr->type : (int)unknown);
  }
}


static void create_meta_body_view_sub2(ph_pt2pt_t pt2pt, char * body) {
  int PH_ENTRY_SIZE = 100;
  ph_msg_cell_t p;
  char * s;
 
  for (p = pt2pt->msg_list->head, s = body; p != NULL; p = p->next, s += PH_ENTRY_SIZE) {
    ph_pthread_mutex_lock(&p->self->mp);
    snprintf(s, PH_ENTRY_SIZE, "%d " PH_LLD " %d %d %d\n", 
	     p->self->type, p->self->dest, p->self->len, p->self->tag, get_conn_no(pt2pt, p->self->to_conn));
    ph_pthread_mutex_unlock(&p->self->mp);
  }
}


static char * create_meta_body_view(ph_pt2pt_t pt2pt, va_list ap, size_t * len_p) {
  int PH_ENTRY_SIZE = 100;
  char * body;
  size_t len;

  ph_pthread_mutex_lock(&pt2pt->routing->mp);
  ph_pthread_mutex_lock(&pt2pt->conn_list->mp);
  ph_pthread_mutex_lock(&pt2pt->msg_list->mp);

  len = 400 + PH_ENTRY_SIZE*(pt2pt->routing->entry_list->nelm + pt2pt->msg_list->nelm);
  (*len_p) = len;
  body = (char *)ph_calloc(len, sizeof(char));
  ph_mzero(body, len);

  create_meta_body_view_sub0(pt2pt, body); 		/* misc */
  create_meta_body_view_sub1(pt2pt, body + 400); 	/* routing */
  create_meta_body_view_sub2(pt2pt, body + 400 + PH_ENTRY_SIZE*(pt2pt->routing->entry_list->nelm)); 	/* messages */
    
  ph_pthread_mutex_unlock(&pt2pt->msg_list->mp);
  ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
  ph_pthread_mutex_unlock(&pt2pt->routing->mp);

  return body;
}

#ifdef PH_ROUTING_ENABLE_DOMINATING

static char * create_meta_body_conn_type(ph_pt2pt_t pt2pt, va_list ap, size_t * len_p) {
  char * body;
  ph_vp_set_t vps[2], local_vps;
  ph_addr_type_t type;

  type = (ph_addr_type_t)va_arg(ap, ph_addr_type_t);
  local_vps = (ph_vp_set_t)va_arg(ap, ph_vp_set_t);
  vps[tcp] = (ph_vp_set_t)va_arg(ap, ph_vp_set_t);
  vps[ssh] = (ph_vp_set_t)va_arg(ap, ph_vp_set_t);
  assert(vps[tcp] != NULL);
  assert(vps[ssh] != NULL);

  (*len_p) = 10 + (ph_vp_range_list_get_nelm(local_vps) + ph_vp_range_list_get_nelm(vps[tcp]) + ph_vp_range_list_get_nelm(vps[ssh])+ 3)*50;
  body = (char *)ph_calloc((*len_p), sizeof(char));
  ph_mzero(body, (*len_p));
  snprintf(body, 10, "%d ", type);

  ph_vp_set_vps2str(local_vps, body+10);
  ph_vp_set_vps2str(vps[tcp], body+10 + (ph_vp_range_list_get_nelm(local_vps) + 1)*50);
  ph_vp_set_vps2str(vps[ssh], body+10 + (ph_vp_range_list_get_nelm(local_vps) + ph_vp_range_list_get_nelm(vps[tcp]) + 2)*50);

  return body;
}
#endif


static char * create_meta_body(ph_pt2pt_t pt2pt, ph_msg_meta_type_t meta_type, va_list ap, size_t * len_p) {
  switch (meta_type) {
  case routing:
    return create_meta_body_routing(pt2pt, ap, len_p);
    break;
  case view:
    return create_meta_body_view(pt2pt, ap, len_p);
    break;
#ifdef PH_ROUTING_ENABLE_DOMINATING
  case conn_type:
  case conn_type_reply:
    return create_meta_body_conn_type(pt2pt, ap, len_p);
    break;
#endif
  default:
    ph_err_quit("create_meta_body() : no matching\n");
    break;
  }
  return NULL;
}


ph_msg_t ph_pt2pt_create_meta_msg(ph_pt2pt_t pt2pt, ph_conn_t conn, ph_msg_meta_type_t meta_type, ...) {
  va_list ap;
  ph_msg_t msg;
  char * body;
  size_t len;
  
  assert(conn != NULL);
  va_start(ap, meta_type);
  body = create_meta_body(pt2pt, meta_type, ap, &len);
  msg = ph_msg_new_meta(conn, meta_type, body, len);
  va_end(ap);
  return msg;
}


void ph_pt2pt_send_meta_msg(ph_pt2pt_t pt2pt, ph_msg_t msg) {
  assert(0);
}


static void handle_meta_msg_routing(ph_pt2pt_t pt2pt, ph_msg_t msg) {
  ph_routing_event_list_enqueue(pt2pt->routing->event_list, 
				ph_routing_event_new(update, msg->from_conn, ph_routing_analyze_msg(msg->body)));
}


#ifdef PH_ROUTING_ENABLE_DOMINATING

static void handle_meta_msg_conn_type_sub(ph_pt2pt_t pt2pt, ph_msg_t msg) {
   ph_addr_type_t type;

  sscanf(msg->body, "%d", (int *)&type);
  if (type != unknown) {
    ph_conn_set_addr_type(msg->from_conn, type);
  }


  {
    ph_routing_t routing = pt2pt->routing;
    ph_routing_neigh_list_t neigh_list;
    ph_routing_neigh_t neigh;

    ph_pthread_mutex_lock(&routing->mp);
    //fprintf(stderr, "accessing neigh_list (start)\n");

    neigh_list = routing->neigh_list;

    /* $B<u$1<h$C$?>pJs$r$b$H$K(Bpt2pt->routing->neigh_list $B$KDI2C$9$k(B */
    if ((neigh = ph_routing_neigh_list_find(neigh_list, &ph_routing_neigh_is_same_conn, msg->from_conn)) == NULL) {
      /* if a new link is added */
      neigh = ph_routing_neigh_new(msg->from_conn);
      ph_routing_neigh_list_insert(neigh_list, neigh);
    }

    
    //neigh->addr = xxxx; /* ????? */
    ph_vp_set_dispose(neigh->local);
    ph_vp_set_dispose(neigh->p_neigh[tcp]);
    ph_vp_set_dispose(neigh->p_neigh[ssh]);

    gettimeofday(&neigh->time, NULL);
    neigh->local = ph_vp_set_str2vps(msg->body+10);
    neigh->p_neigh[tcp] = ph_vp_set_str2vps(msg->body+10 + 
					    (ph_vp_range_list_get_nelm(neigh->local) + 1)*50);
    neigh->p_neigh[ssh] = ph_vp_set_str2vps(msg->body+10 +
					    (ph_vp_range_list_get_nelm(neigh->local) + ph_vp_range_list_get_nelm(neigh->p_neigh[tcp])+ 2)*50);

    //fprintf(stderr, "accessing neigh_list (end)\n");
    ph_pthread_mutex_unlock(&routing->mp);
  }
}


static void handle_meta_msg_conn_type(ph_pt2pt_t pt2pt, ph_msg_t msg) {
  handle_meta_msg_conn_type_sub(pt2pt, msg);

 {
    ph_vp_set_t vps[2], local_vps;
    ph_routing_neigh_cell_t q;
    
    local_vps = ph_vp_set_sync_dup(pt2pt->vp_set);
    
    ph_pthread_mutex_lock(&pt2pt->routing->mp);
    ph_pthread_mutex_lock(&pt2pt->conn_list->mp);
    
    vps[tcp] = ph_vp_set_make_empty();
  vps[ssh] = ph_vp_set_make_empty();
  for (q = pt2pt->routing->neigh_list->head; q != NULL; q = q->next) {
    ph_vp_set_t tmp;

    if (q->self->conn == NULL) {
      continue;
    }
    switch (ph_conn_get_addr_type(q->self->conn)) {
    case tcp:
    case unknown:
      tmp = ph_vp_set_make_union(vps[tcp], q->self->local);
      ph_vp_set_dispose(vps[tcp]);
      vps[tcp] = tmp;
      break;
    case ssh:
      tmp = ph_vp_set_make_union(vps[ssh], q->self->local);
      ph_vp_set_dispose(vps[ssh]);
      vps[ssh] = tmp;
      break;
    default:
      ph_err_quit("connector_check() : no matching");
      break;
    }
  }
  ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
  ph_pthread_mutex_unlock(&pt2pt->routing->mp);


  ph_pt2pt_handle_msg(pt2pt, ph_pt2pt_create_meta_msg(pt2pt, msg->from_conn, conn_type_reply, ph_conn_get_addr_type(msg->from_conn), local_vps, vps[tcp], vps[ssh]));

  ph_vp_set_dispose(local_vps);
  ph_vp_set_dispose(vps[tcp]);
  ph_vp_set_dispose(vps[ssh]);   
 }

}


static void handle_meta_msg_conn_type_reply(ph_pt2pt_t pt2pt, ph_msg_t msg) {
  handle_meta_msg_conn_type_sub(pt2pt, msg);
}

#endif

void ph_pt2pt_recv_meta_msg(ph_pt2pt_t pt2pt, ph_msg_t msg) {
  assert(msg != NULL);
  assert(msg->type == meta);
  
  switch (msg->meta_type) {
  case routing:
    handle_meta_msg_routing(pt2pt, msg);
    break;
#ifdef PH_ROUTING_ENABLE_DOMINATING
  case conn_type:
    handle_meta_msg_conn_type(pt2pt, msg);
    break;
  case conn_type_reply:
    handle_meta_msg_conn_type_reply(pt2pt, msg);
    break;
#endif
  default:
    ph_err_quit("ph_pt2pt_recv_msg() : no matching");
    break;
  }
  
  ph_msg_dispose(msg);
}










