/**************************************
 * $Author: kaneda $
 * $Date: 2002/12/20 10:14:42 $
 **************************************/

/* task.c */

#include "task.h"


task_t task_new() {
  task_t task;

  task = (task_t)Malloc(sizeof(struct task));

  task->argv = task->in_argv = task->out_argv = NULL;
  task->id = task->in_id = task->out_id = 0LL;
  task->pid = 0;

  return task;
}


void argv_dispose(char ** argv) {
  int i;

  if (argv == NULL) {
    err_msg("argv_dispose() : NULL input");
    return;
  }
  
  for (i = 0; argv[i] != NULL; i++) {
    Free(argv[i]);
  }
  Free(argv);
}


void task_dispose(task_t task) { 
  if (task == NULL) {
    err_msg("task_dispose() : NULL input");
    return;
  }

  argv_dispose(task->argv);
  Free(task);
}


int task_argv2str(task_t task, char * s, size_t len) {
  int i, n;

  assert(task->argv != NULL);

  Mzero(s, len);

  n = 0;
  for (i = 0; task->argv[i] != NULL; i++) {   
    int m = Strlen(task->argv[i])+1;

    Mmove(s + n, task->argv[i], m);
    n += m;
    s[n-1] = '\0';
    assert(n <= len);
  }

  return n;
}


void task_set_argv(task_t task, char * s, int argn) {
  int i;
  char * t;

  task->argv = (char **)Calloc(argn+1, sizeof(char*));
  
  t = s;
  for (i = 0; i < argn; i++) {
    task->argv[i] = Mdup(t, Strlen(t)+1);
    t += Strlen(t)+1;
  }

  for (i = 0; i < argn; i++) {
    debug_print("task_set_argv()\t\t\t: argv[%o]=\"%s\"\n", i, task->argv[i]);
  }

  task->argv[argn] = NULL;
}


DEFINE_LIST_FUNC_BODY(task)





static char * print_state(task_state_t state) {
  switch (state) {
  case TASK_EXECUTING:
    return "executing";
    break;
  case TASK_WAITING:
    return "waiting";
    break;
  case TASK_DONE:
    return "done";
    break;
  default:
    err_quit("state_print() : no matching (state=%d)", state);
    break;
  }
  return NULL;
}

static void task_print(task_t task, void * arg) {
  debug_print("(%p) vp=%lld, id=%lld, argv[0]=%s, in_vp=%lld, in_id=%lld, in_argv[0]=%s, out_vp=%lld, out_id=%lld, out_argv[0]=%s, pid=%d \n",
	      task,
	      task->vp,
	      task->id,
	      (task->argv != NULL) ? task->argv[0] : "null",
	      task->in_vp,
	      task->in_id,
	      (task->in_argv != NULL) ? task->in_argv[0] : "null",
	      task->out_vp,
	      task->out_id,
	      (task->out_argv != NULL) ? task->out_argv[0] : "null",
	      (int)task->pid);
}

void task_list_print(task_list_t task_list) {
 assert(task_list != NULL);

  if (task_list_is_empty(task_list)) {
     debug_print2("[task_list]\n");
  } else {
    debug_print2("[task_list \n");
    task_list_map(task_list, &task_print, NULL);
    debug_print2("]\n");
  }
}



static int is_same_pid(task_t task, void * arg) {
  pid_t pid = (pid_t)arg;
  return (task->pid == pid);
}

task_t task_list_find_same_pid(task_list_t task_list, pid_t pid) {
  return task_list_find(task_list, &is_same_pid, (void *)pid);
}


task_cell_t task_list_init_loop(task_list_t task_list) {
  assert(task_list != NULL);
  return task_list->head;
}


task_cell_t task_list_iterate(task_list_t task_list, task_cell_t p) {
  task_cell_t next;
  
  assert(task_list != NULL);
  assert(p != NULL);

  next = p->next;
  if (p->self->state == TASK_DONE) {
    task_list_delete(task_list, p->self);
  }
  return next;
}










